import string
import pytest

from minictl import CTLFormula as CF


class TestCreationErrors:
    @pytest.mark.parametrize("variable_name", string.ascii_uppercase)
    def test_variable_lowercase(self, variable_name):
        with pytest.raises(ValueError):
            CF(variable_name)

    @pytest.mark.parametrize("formula_type", ["p", "q", "TOP", "BOT"])
    def test_no_args(self, formula_type):
        for i in range(1, 100):
            with pytest.raises(ValueError):
                CF(formula_type, *[CF("q") for _ in range(i)])

    @pytest.mark.parametrize(
        "formula_type", ["Neg", "EX", "AX", "EF", "AF", "EG", "AG"]
    )
    def test_one_args(self, formula_type):
        for i in range(2, 100):
            with pytest.raises(ValueError):
                CF(formula_type, *[CF("q") for _ in range(i)])
        with pytest.raises(ValueError):
            CF(formula_type)
        with pytest.raises(TypeError):
            CF(formula_type, "Something that is not an LTLFormula")

    @pytest.mark.parametrize(
        "formula_type",
        ["And", "Or", "ImpliesR", "ImpliesL", "BiImplies", "EU", "AU"],
    )
    def test_two_args(self, formula_type):
        for i in range(3, 100):
            with pytest.raises(ValueError):
                CF(formula_type, *[CF("q") for _ in range(i)])
        for i in range(2):
            with pytest.raises(ValueError):
                CF(formula_type, *[CF("q") for _ in range(i)])
        with pytest.raises(TypeError):
            CF(formula_type, "Something that is", "not an LTLFormula")


class TestParse:
    def test_var(self):
        assert CF("p") == CF.parse("p")

    def test_top(self):
        assert CF("TOP") == CF.parse("TOP")

    def test_bot(self):
        assert CF("BOT") == CF.parse("BOT")

    def test_neg(self):
        assert CF("Neg", CF("p")) == CF.parse("!p")

    def test_ex(self):
        assert CF("EX", CF("p")) == CF.parse("EXp")

    def test_ax(self):
        assert CF("AX", CF("p")) == CF.parse("AXp")

    def test_ef(self):
        assert CF("EF", CF("p")) == CF.parse("EFp")

    def test_af(self):
        assert CF("AF", CF("p")) == CF.parse("AFp")

    def test_eg(self):
        assert CF("EG", CF("p")) == CF.parse("EGp")

    def test_ag(self):
        assert CF("AG", CF("p")) == CF.parse("AGp")

    def test_and(self):
        assert CF("And", CF("p"), CF("q")) == CF.parse("p and q")

    def test_or(self):
        assert CF("Or", CF("p"), CF("q")) == CF.parse("p or q")

    def test_implies_right(self):
        assert CF("ImpliesR", CF("p"), CF("q")) == CF.parse("p -> q")

    def test_implies_left(self):
        assert CF("ImpliesL", CF("p"), CF("q")) == CF.parse("p <- q")

    def test_implies_bi(self):
        assert CF("BiImplies", CF("p"), CF("q")) == CF.parse("p <-> q")

    def test_eu(self):
        assert CF("EU", CF("p"), CF("q")) == CF.parse("E[pUq]")

    def test_au(self):
        assert CF("AU", CF("p"), CF("q")) == CF.parse("A[pUq]")


class TestSymbols:
    def test_top(self):
        assert str(CF.parse("TOP")) == "⊤"

    def test_bot(self):
        assert str(CF.parse("BOT")) == "⊥"

    def test_neg(self):
        assert str(CF.parse("!p")) == "¬(p)"

    def test_ex(self):
        assert str(CF.parse("EXp")) == "EX(p)"

    def test_ax(self):
        assert str(CF.parse("AXp")) == "AX(p)"

    def test_ef(self):
        assert str(CF.parse("EFp")) == "EF(p)"

    def test_af(self):
        assert str(CF.parse("AFp")) == "AF(p)"

    def test_eg(self):
        assert str(CF.parse("EGp")) == "EG(p)"

    def test_ag(self):
        assert str(CF.parse("AGp")) == "AG(p)"

    def test_and(self):
        assert str(CF.parse("p and q")) == "(p)∧(q)"

    def test_or(self):
        assert str(CF.parse("p or q")) == "(p)∧(q)"

    def test_implies_right(self):
        assert str(CF.parse("p -> q")) == "(p)→(q)"

    def test_implies_left(self):
        assert str(CF.parse("p <- q")) == "(p)←(q)"

    def test_implies_bi(self):
        assert str(CF.parse("p <-> q")) == "(p)↔(q)"

    def test_eu(self):
        assert str(CF.parse("E[pUq]")) == "E[(p)U(q)]"

    def test_au(self):
        assert str(CF.parse("A[pUq]")) == "A[(p)U(q)]"

    def test_complex_1(self):
        assert (
            str(CF.parse("E[pU(q and p)] -> AXz or EG(A[rUw])"))
            == "(E[(p)U((q)∧(p))])→((AX(z))∧(EG(A[(r)U(w)])))"
        )

    def test_complex_2(self):
        assert (
            str(CF.parse("!BOT -> !AG(p <-> q or A[zUw])"))
            == "(¬(⊥))→(¬(AG((p)↔((q)∧(A[(z)U(w)])))))"
        )


class TestEq:
    def test_implies_right(self):
        assert CF.parse("(p) -> q") == CF.parse("p -> (q)") == CF.parse("(p -> q)")

    def test_implies_left(self):
        assert CF.parse("(p) <- q") == CF.parse("p <- (q)") == CF.parse("(p <- q)")

    def test_complex(self):
        assert (
            CF.parse("(!TOP -> A[(AXq)U(EFp)]) <-> TOP")
            == CF.parse("((!(TOP) -> A[(AX(q))U(EF(p))]) <-> (TOP))")
            == CF.parse("(!TOP -> A[AXq U EFp]) <-> TOP")
        )



from minictl import CTLFormula, CTLChecker, State, Model


class TestCheckerBasics:
    s1 = State("s1", {"p"})
    s2 = State("s2", {"p", "q"})
    s3 = State("s3", {"p", "q"})
    s4 = State("s4", set())
    s5 = State("s5", {"q"})
    s6 = State("s6", {"q"})
    model = Model(
        [s1, s2, s3, s4, s5, s6],
        {
            "s1": ["s1", "s2", "s3"],
            "s2": ["s1", "s2", "s3"],
            "s3": ["s4", "s5"],
            "s4": ["s1", "s6"],
            "s5": ["s4", "s6"],
            "s6": ["s1", "s2"],
        },
    )

    def test_topbot(self):
        checker = CTLChecker(self.model)
        assert checker.check(CTLFormula.parse("TOP")) == {
            "s1",
            "s2",
            "s3",
            "s4",
            "s5",
            "s6",
        }
        assert checker.check(CTLFormula.parse("BOT")) == set()

    def test_vars(self):
        checker = CTLChecker(self.model)
        assert checker.check(CTLFormula.parse("p")) == {"s1", "s2", "s3"}
        assert checker.check(CTLFormula.parse("q")) == {"s2", "s3", "s5", "s6"}

    def test_neg(self):
        checker = CTLChecker(self.model)
        assert checker.check(CTLFormula.parse("!p")) == {"s4", "s5", "s6"}
        assert checker.check(CTLFormula.parse("!q")) == {"s1", "s4"}

    def test_and(self):
        checker = CTLChecker(self.model)
        assert checker.check(CTLFormula.parse("p and q")) == {"s2", "s3"}

    def test_or(self):
        checker = CTLChecker(self.model)
        assert checker.check(CTLFormula.parse("p or q")) == {
            "s1",
            "s2",
            "s3",
            "s5",
            "s6",
        }

    def test_implies_r(self):
        checker = CTLChecker(self.model)
        assert checker.check(CTLFormula.parse("p -> q")) == {
            "s2",
            "s3",
            "s4",
            "s5",
            "s6",
        }

    def test_implies_l(self):
        checker = CTLChecker(self.model)
        assert checker.check(CTLFormula.parse("p <- q")) == {"s1", "s2", "s3", "s4"}

    def test_implies_bi(self):
        checker = CTLChecker(self.model)
        assert checker.check(CTLFormula.parse("p <-> q")) == {"s2", "s3", "s4"}

    def test_ax(self):
        checker = CTLChecker(self.model)
        assert checker.check(CTLFormula.parse("AXp")) == {"s1", "s2", "s6"}
        assert checker.check(CTLFormula.parse("AXq")) == set()

    def test_ex(self):
        checker = CTLChecker(self.model)
        assert checker.check(CTLFormula.parse("EXp")) == {"s1", "s2", "s4", "s6"}
        assert checker.check(CTLFormula.parse("EXq")) == {
            "s1",
            "s2",
            "s3",
            "s4",
            "s5",
            "s6",
        }

    def test_au(self):
        checker = CTLChecker(self.model)
        assert checker.check(CTLFormula.parse("A[qUp]")) == {"s1", "s2", "s3", "s6"}
        assert checker.check(CTLFormula.parse("A[pUq]")) == {
            "s1",
            "s2",
            "s3",
            "s4",
            "s5",
            "s6",
        }

    def test_eu(self):
        checker = CTLChecker(self.model)
        assert checker.check(CTLFormula.parse("E[pUq]")) == {
            "s1",
            "s2",
            "s3",
            "s5",
            "s6",
        }
        assert checker.check(CTLFormula.parse("E[qUp]")) == {
            "s1",
            "s2",
            "s3",
            "s5",
            "s6",
        }

    def test_af(self):
        checker = CTLChecker(self.model)
        assert checker.check(CTLFormula.parse("AFq")) == {
            "s2",
            "s3",
            "s5",
            "s6",
        }
        assert checker.check(CTLFormula.parse("AFp")) == {
            "s1",
            "s2",
            "s3",
            "s4",
            "s5",
            "s6",
        }

    def test_ef(self):
        checker = CTLChecker(self.model)
        assert checker.check(CTLFormula.parse("EFp")) == {
            "s1",
            "s2",
            "s3",
            "s4",
            "s5",
            "s6",
        }
        assert checker.check(CTLFormula.parse("EFq")) == {
            "s1",
            "s2",
            "s3",
            "s4",
            "s5",
            "s6",
        }

    def test_ag(self):
        checker = CTLChecker(self.model)
        assert checker.check(CTLFormula.parse("AGp")) == set()
        assert checker.check(CTLFormula.parse("AGq")) == set()

    def test_eg(self):
        checker = CTLChecker(self.model)
        assert checker.check(CTLFormula.parse("EGp")) == {"s1", "s2"}
        assert checker.check(CTLFormula.parse("EGq")) == {"s2", "s3", "s5", "s6"}


# These come from lecture and workgroup slides of the course.
# I thought making it do all assignments would be a good and funny test. :P
class TestCheckerFormulasLectures:
    s1 = State("s1", {"p"})
    s2 = State("s2", {"p", "q"})
    s3 = State("s3", {"p", "q"})
    s4 = State("s4", set())
    s5 = State("s5", {"q"})
    s6 = State("s6", {"q"})
    model = Model(
        [s1, s2, s3, s4, s5, s6],
        {
            "s1": ["s1", "s2", "s3"],
            "s2": ["s1", "s2", "s3"],
            "s3": ["s4", "s5"],
            "s4": ["s1", "s6"],
            "s5": ["s4", "s6"],
            "s6": ["s1", "s2"],
        },
    )

    def test1(self):
        checker = CTLChecker(self.model)
        assert checker.check(CTLFormula.parse("E[!pU!q]")) == {"s1", "s4", "s5", "s6"}

    def test2(self):
        checker = CTLChecker(self.model)
        assert checker.check(CTLFormula.parse("E[pU!q]")) == {"s1", "s2", "s3", "s4"}

    def test3(self):
        checker = CTLChecker(self.model)
        assert checker.check(CTLFormula.parse("A[pU(EXq)]")) == {
            "s1",
            "s2",
            "s3",
            "s4",
            "s5",
            "s6",
        }

    def test4(self):
        checker = CTLChecker(self.model)
        assert checker.check(CTLFormula.parse("AX EXp")) == {"s4", "s5", "s6"}

    def test5(self):
        checker = CTLChecker(self.model)
        assert checker.check(CTLFormula.parse("AF EGp")) == {
            "s1",
            "s2",
            "s3",
            "s4",
            "s5",
            "s6",
        }


class TestCheckerFormulasMutualExclusion:
    s0 = State("s0", {"n1", "n2"})
    s1 = State("s1", {"t1", "n2"})
    s2 = State("s2", {"c1", "n2"})
    s3 = State("s3", {"t1", "t2"})
    s4 = State("s4", {"c1", "t2"})
    s5 = State("s5", {"n1", "t2"})
    s6 = State("s6", {"n1", "c2"})
    s7 = State("s7", {"t1", "c2"})
    s8 = State("s8", {"t1", "t2"})

    model = Model(
        [s0, s1, s2, s3, s4, s5, s6, s7, s8],
        {
            "s0": ["s1", "s5"],
            "s1": ["s3", "s2"],
            "s2": ["s4", "s5"],
            "s3": ["s4"],
            "s4": ["s5"],
            "s5": ["s6", "s8"],
            "s6": ["s1", "s7"],
            "s7": ["s1"],
            "s8": ["s7"],
        },
    )

    def test_safety(self):
        checker = CTLChecker(self.model)
        assert checker.check(CTLFormula.parse("AG!(c1 and c2)")) == {
            "s0",
            "s1",
            "s2",
            "s3",
            "s4",
            "s5",
            "s6",
            "s7",
            "s8",
        }

    def test_liveness(self):
        checker = CTLChecker(self.model)
        assert checker.check(CTLFormula.parse("AG(t1 -> AFc1)")) == {
            "s0",
            "s1",
            "s2",
            "s3",
            "s4",
            "s5",
            "s6",
            "s7",
            "s8",
        }
        assert checker.check(CTLFormula.parse("AG(t2 -> AFc2)")) == {
            "s0",
            "s1",
            "s2",
            "s3",
            "s4",
            "s5",
            "s6",
            "s7",
            "s8",
        }



import string
import pytest

from minictl import LTLFormula as LF


class TestCreationErrors:
    @pytest.mark.parametrize("variable_name", string.ascii_uppercase)
    def test_variable_lowercase(self, variable_name):
        with pytest.raises(ValueError):
            LF(variable_name)

    @pytest.mark.parametrize("formula_type", ["p", "q", "TOP", "BOT"])
    def test_no_args(self, formula_type):
        for i in range(1, 100):
            with pytest.raises(ValueError):
                LF(formula_type, *[LF("q") for _ in range(i)])

    @pytest.mark.parametrize("formula_type", ["Neg", "X", "F", "G"])
    def test_one_args(self, formula_type):
        for i in range(2, 100):
            with pytest.raises(ValueError):
                LF(formula_type, *[LF("q") for _ in range(i)])
        with pytest.raises(ValueError):
            LF(formula_type)
        with pytest.raises(TypeError):
            LF(formula_type, "Something that is not an LTLFormula")

    @pytest.mark.parametrize(
        "formula_type",
        ["And", "Or", "ImpliesR", "ImpliesL", "BiImplies", "U", "W", "R"],
    )
    def test_two_args(self, formula_type):
        for i in range(3, 100):
            with pytest.raises(ValueError):
                LF(formula_type, *[LF("q") for _ in range(i)])
        for i in range(2):
            with pytest.raises(ValueError):
                LF(formula_type, *[LF("q") for _ in range(i)])
        with pytest.raises(TypeError):
            LF(formula_type, "Something that is", "not an LTLFormula")


class TestParse:
    def test_var(self):
        assert LF("p") == LF.parse("p")

    def test_top(self):
        assert LF("TOP") == LF.parse("TOP")

    def test_bot(self):
        assert LF("BOT") == LF.parse("BOT")

    def test_neg(self):
        assert LF("Neg", LF("p")) == LF.parse("!p")

    def test_x(self):
        assert LF("X", LF("p")) == LF.parse("Xp")

    def test_f(self):
        assert LF("F", LF("p")) == LF.parse("Fp")

    def test_g(self):
        assert LF("G", LF("p")) == LF.parse("Gp")

    def test_and(self):
        assert LF("And", LF("p"), LF("q")) == LF.parse("p and q")

    def test_or(self):
        assert LF("Or", LF("p"), LF("q")) == LF.parse("p or q")

    def test_implies_right(self):
        assert LF("ImpliesR", LF("p"), LF("q")) == LF.parse("p -> q")

    def test_implies_left(self):
        assert LF("ImpliesL", LF("p"), LF("q")) == LF.parse("p <- q")

    def test_implies_bi(self):
        assert LF("BiImplies", LF("p"), LF("q")) == LF.parse("p <-> q")

    def test_u(self):
        assert LF("U", LF("p"), LF("q")) == LF.parse("pUq")

    def test_w(self):
        assert LF("W", LF("p"), LF("q")) == LF.parse("pWq")

    def test_r(self):
        assert LF("R", LF("p"), LF("q")) == LF.parse("pRq")


class TestSymbols:
    def test_top(self):
        assert str(LF.parse("TOP")) == "⊤"

    def test_bot(self):
        assert str(LF.parse("BOT")) == "⊥"

    def test_neg(self):
        assert str(LF.parse("!p")) == "¬(p)"

    def test_x(self):
        assert str(LF.parse("Xp")) == "X(p)"

    def test_f(self):
        assert str(LF.parse("Fp")) == "F(p)"

    def test_g(self):
        assert str(LF.parse("Gp")) == "G(p)"

    def test_and(self):
        assert str(LF.parse("p and q")) == "(p)∧(q)"

    def test_or(self):
        assert str(LF.parse("p or q")) == "(p)∧(q)"

    def test_implies_right(self):
        assert str(LF.parse("p -> q")) == "(p)→(q)"

    def test_implies_left(self):
        assert str(LF.parse("p <- q")) == "(p)←(q)"

    def test_implies_bi(self):
        assert str(LF.parse("p <-> q")) == "(p)↔(q)"

    def test_u(self):
        assert str(LF.parse("pUq")) == "(p)U(q)"

    def test_w(self):
        assert str(LF.parse("pWq")) == "(p)W(q)"

    def test_r(self):
        assert str(LF.parse("pRq")) == "(p)R(q)"

    def test_complex_1(self):
        assert (
            str(LF.parse("pU(q and p) -> Xz or G(rUw)"))
            == "((p)U((q)∧(p)))→((X(z))∧(G((r)U(w))))"
        )

    def test_complex_2(self):
        assert (
            str(LF.parse("!BOT -> !G(p <-> q or zWw)"))
            == "(¬(⊥))→(¬(G((p)↔((q)∧((z)W(w))))))"
        )


class TestEq:
    def test_implies_right(self):
        assert LF.parse("(p) -> q") == LF.parse("p -> (q)") == LF.parse("(p -> q)")

    def test_implies_left(self):
        assert LF.parse("(p) <- q") == LF.parse("p <- (q)") == LF.parse("(p <- q)")

    def test_complex(self):
        assert (
            LF.parse("(!TOP -> (Xq)U(Fp)) <-> TOP")
            == LF.parse("((!(TOP) -> (X(q))U(F(p))) <-> (TOP))")
            == LF.parse("(!TOP -> Xq U Fp) <-> TOP")
        )



from copy import copy

import pytest

from minictl import CTLFormula, CTLChecker, Model, State


def ef(states: set[str], model: Model) -> set[str]:
    while True:
        next_states = copy(states)

        for s in model.all():
            reachables = model.get_next(s)
            if reachables.intersection(states):
                next_states.add(s)

        if next_states == states:
            return states
        else:
            states = next_states


def eu(lhs: set[str], rhs: set[str], model: Model) -> set[str]:
    states = rhs
    while True:
        next_states = copy(states)

        for s in model.all():
            if s in lhs:
                reachables = model.get_next(s)
                if reachables.intersection(states):
                    next_states.add(s)

        if next_states == states:
            return states
        else:
            states = next_states


def empty(states: set[str], model: Model) -> set[str]:
    return set()


class TestModularChecker:
    s1 = State("s1", {"p"})
    s2 = State("s2", {"p", "q"})
    s3 = State("s3", {"p", "q"})
    s4 = State("s4", set())
    s5 = State("s5", {"q"})
    s6 = State("s6", {"q"})
    model = Model(
        [s1, s2, s3, s4, s5, s6],
        {
            "s1": ["s1", "s2", "s3"],
            "s2": ["s1", "s2", "s3"],
            "s3": ["s4", "s5"],
            "s4": ["s1", "s6"],
            "s5": ["s4", "s6"],
            "s6": ["s1", "s2"],
        },
    )

    def test_ef_correct(self):
        checker = CTLChecker(self.model)
        checker.set_custom("EF", ef)
        assert checker.check(CTLFormula.parse("EFp")) == {
            "s1",
            "s2",
            "s3",
            "s4",
            "s5",
            "s6",
        }
        assert checker.check(CTLFormula.parse("EFq")) == {
            "s1",
            "s2",
            "s3",
            "s4",
            "s5",
            "s6",
        }

    def test_ef_correct_debug(self):
        checker = CTLChecker(self.model)
        checker.set_custom("EF", ef)
        assert checker.check(CTLFormula.parse("EFp"), debug=True) == {
            "s1",
            "s2",
            "s3",
            "s4",
            "s5",
            "s6",
        }
        assert checker.check(CTLFormula.parse("EFq"), debug=True) == {
            "s1",
            "s2",
            "s3",
            "s4",
            "s5",
            "s6",
        }

    def test_ef_incorrect(self):
        checker = CTLChecker(self.model)
        checker.set_custom("EF", empty)
        assert checker.check(CTLFormula.parse("EFp")) == set()
        assert checker.check(CTLFormula.parse("EFq")) == set()

    def test_ef_incorrect_debug(self):
        checker = CTLChecker(self.model)
        checker.set_custom("EF", empty)
        with pytest.raises(RuntimeError):
            assert checker.check(CTLFormula.parse("EFp"), debug=True)
        with pytest.raises(RuntimeError):
            assert checker.check(CTLFormula.parse("EFq"), debug=True)

    def test_eu_correct(self):
        checker = CTLChecker(self.model)
        checker.set_custom("EU", eu)
        assert checker.check(CTLFormula.parse("E[pUq]")) == {
            "s1",
            "s2",
            "s3",
            "s5",
            "s6",
        }
        assert checker.check(CTLFormula.parse("E[qUp]")) == {
            "s1",
            "s2",
            "s3",
            "s5",
            "s6",
        }

    def test_eu_correct_debug(self):
        checker = CTLChecker(self.model)
        checker.set_custom("EU", eu)
        assert checker.check(CTLFormula.parse("E[pUq]"), debug=True) == {
            "s1",
            "s2",
            "s3",
            "s5",
            "s6",
        }
        assert checker.check(CTLFormula.parse("E[qUp]"), debug=True) == {
            "s1",
            "s2",
            "s3",
            "s5",
            "s6",
        }

    def test_is_modified(self):
        checker = CTLChecker(self.model)
        assert not checker.is_modified()
        checker.set_custom("EF", ef)
        assert checker.is_modified()
        checker.set_custom("EF", empty)
        assert checker.is_modified()

    def test_get_model(self):
        checker = CTLChecker(self.model)
        model = checker.get_model()
        assert model.all() == {"s1", "s2", "s3", "s4", "s5", "s6"}

    def test_cannot_set_after_check(self):
        checker = CTLChecker(self.model)
        checker.set_custom("EF", ef)
        checker.check(CTLFormula.parse("EFp"), debug=True)
        with pytest.raises(ValueError):
            checker.set_custom("EU", eu)



from minictl import State


class TestState:
    def test_creation(self):
        state = State("s1", {"p", "q"})
        assert state.name == "s1"
        assert state.variables == {"p", "q"}

    def test_contains(self):
        state = State("s1", {"p", "q"})
        assert state.contains("p")
        assert state.contains("q")
        assert not state.contains("r")
        assert not state.contains("s1")



import pytest
from minictl import Model, State


class TestModelBasics:
    s1 = State("s1", {"p", "q"})
    s2 = State("s2", set())
    s3 = State("s3", {"q"})

    def test_getstate(self):
        model = Model([self.s3], {"s3": ["s3"]})
        instate = model.get_state("s3")
        assert instate.name == "s3"
        assert instate.contains("q")
        assert not instate.contains("p")

    def test_all(self):
        model = Model([self.s3], {"s3": ["s3"]})
        assert model.all() == {"s3"}
        model = Model(
            [self.s1, self.s2, self.s3], {"s1": ["s1"], "s2": ["s2"], "s3": ["s3"]}
        )
        assert model.all() == {"s1", "s2", "s3"}

    def test_all_containing(self):
        model = Model(
            [self.s1, self.s2, self.s3], {"s1": ["s1"], "s2": ["s2"], "s3": ["s3"]}
        )
        assert model.all_containing("p") == {"s1"}
        assert model.all_containing("q") == {"s1", "s3"}

    def test_all_except(self):
        model = Model(
            [self.s1, self.s2, self.s3], {"s1": ["s1"], "s2": ["s2"], "s3": ["s3"]}
        )
        assert model.all_except(set()) == {"s1", "s2", "s3"}
        assert model.all_except({"s2"}) == {"s1", "s3"}
        assert model.all_except({"s1", "s2", "s3"}) == set()


class TestModelCreationErrors:
    s1 = State("s1", {"p", "q"})
    s2 = State("s2", set())

    def test_state_not_mentionned(self):
        with pytest.raises(ValueError):
            Model([self.s1, self.s2], {"s1": ["s1"]})

    def test_empty_edge_list(self):
        with pytest.raises(ValueError):
            Model([self.s1, self.s2], {"s1": ["s1"], "s2": []})

    def test_unused_edge(self):
        with pytest.raises(ValueError):
            Model([self.s1], {"s1": ["s1"], "s2": ["s1"]})

    def test_dangeling_edge(self):
        with pytest.raises(ValueError):
            Model([self.s1], {"s1": ["s1", "s2"]})



from typing import Callable, List, Dict, Set

def hello_world() -> str: ...

class LTLFormula:
    name: str
    arguments: List[LTLFormula]
    def __init__(self, name: str, *args: LTLFormula) -> None: ...
    @staticmethod
    def parse(formula: str) -> LTLFormula: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __str__(self) -> str: ...
    def __hash__(self) -> int: ...

class CTLFormula:
    name: str
    arguments: List[CTLFormula]
    def __init__(self, name: str, *args: CTLFormula) -> None: ...
    @staticmethod
    def parse(formula: str) -> CTLFormula: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __str__(self) -> str: ...
    def __hash__(self) -> int: ...

class State:
    name: str
    variables: Set[str]
    def __init__(self, name: str, variables: Set[str]): ...
    def contains(self, var: str) -> bool: ...

class Model:
    def __init__(self, states: List[State], edges: Dict[str, list[str]]) -> None: ...
    def get_state(self, which: str) -> State: ...
    def get_states(self) -> List[State]: ...
    def all(self) -> Set[str]: ...
    def all_containing(self, var: str) -> Set[str]: ...
    def all_except(self, names: Set[str]) -> set[str]: ...
    def pre_e(self, names: Set[str]) -> set[str]: ...
    def pre_a(self, names: Set[str]) -> set[str]: ...
    def get_next(self, name: str) -> Set[str]: ...

# fmt: off
class CTLChecker:
    def __init__(self, model: Model) -> None: ...
    def check(self, formula: CTLFormula, debug: bool = False) -> Set[str]: ...
    def is_modified(self) -> bool: ...
    def get_model(self) -> Model: ...
    def set_custom(
        self,
        target: str,
        func: Callable[[Set[str], Model], set[str]] | Callable[[set[str], set[str], Model], set[str]],
    ) -> None: ...
